# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models
"""

from .core import Base, uuid_column, getset_factory, Setting, Change
from .contact import PhoneNumber, EmailAddress

from .people import Person, PersonPhoneNumber, PersonEmailAddress, PersonMailingAddress
from .users import Role, Permission, User, UserRole
from .stores import Store, StorePhoneNumber, StoreEmailAddress
from .customers import (Customer, CustomerPhoneNumber, CustomerEmailAddress,
                        CustomerGroup, CustomerGroupAssignment, CustomerPerson)

from .org import Department, Subdepartment, Category, Family, ReportCode, DepositLink
from .employees import (Employee, EmployeePhoneNumber, EmployeeEmailAddress,
                        EmployeeStore, EmployeeDepartment)

from .vendors import Vendor, VendorPhoneNumber, VendorEmailAddress, VendorContact
from .products import Brand, Tax, Product, ProductCode, ProductCost, ProductPrice

from .messages import Message, MessageRecipient

from .datasync import DataSyncChange
from .batches import Batch, BatchColumn, BatchRow
from .labels import LabelProfile
from .bouncer import EmailBounce

from rattail.db.batch.model import (BatchMixin, FileBatchMixin,
                                    BatchRowMixin, ProductBatchRowMixin)
from rattail.db.batch.vendorcatalog.model import VendorCatalog, VendorCatalogRow
from rattail.db.batch.vendorinvoice.model import VendorInvoice, VendorInvoiceRow



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Email Framework
"""

from __future__ import unicode_literals, absolute_import

import smtplib
import logging
from email.message import Message
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from mako.template import Template
from mako.lookup import TemplateLookup
from mako.exceptions import TopLevelLookupException

from rattail import exceptions
from rattail.core import UNSPECIFIED
from rattail.files import resource_path
from rattail.util import import_module_path


log = logging.getLogger(__name__)


def send_email(config, key, data={}, attachments=[], fallback_key=None):
    """
    Send an email message of the given type, per config, with the given data
    and/or attachments.
    """
    email = get_email(config, key, fallback_key)
    if email.get_enabled():
        msg = email.make_message(data, attachments)
        deliver_message(config, msg)
    else:
        log.debug("skipping email of type '{0}' per config".format(key))


def get_email(config, key, fallback_key=None):
    """
    Return an email instance of the given type.
    """
    for email in iter_emails(config):
        if email.key == key or email.__name__ == key:
            return email(config, key, fallback_key)
    return Email(config, key, fallback_key)


def iter_emails(config):
    """
    Iterate over all available email types.
    """
    for module in config.getlist('rattail.mail', 'emails', default=[]):
        module = import_module_path(module)
        for name in dir(module):
            obj = getattr(module, name)
            if (isinstance(obj, type) and issubclass(obj, Email)
                and not obj.abstract and obj is not Email):
                yield obj


def deliver_message(config, msg, recipients=UNSPECIFIED):
    """
    Deliver an email message using the given SMTP configuration.
    """
    if recipients is UNSPECIFIED:
        recips = set()
        to = msg.get_all('To')
        if to:
            recips = recips.union(set(to))
        cc = msg.get_all('Cc')
        if cc:
            recips = recips.union(set(cc))
        bcc = msg.get_all('Bcc')
        if bcc:
            recips = recips.union(set(bcc))
    else:
        recips = set(recipients)
    if not recips:
        raise RuntimeError("No recipients for email: {0}".format(repr(msg)))

    server = config.get('rattail.mail', 'smtp.server', default='localhost')
    username = config.get('rattail.mail', 'smtp.username')
    password = config.get('rattail.mail', 'smtp.password')

    if config.getbool('rattail.mail', 'send_emails', default=True):

        log.debug("connecting to server: {0}".format(server))
        session = smtplib.SMTP(server)
        if username and password:
            result = session.login(username, password)
            log.debug("login result is: {0}".format(repr(result)))

        result = session.sendmail(msg['From'], recips, msg.as_string())
        log.debug("sendmail result is: {0}".format(repr(result)))
        session.quit()

    else:
        log.debug("config says no emails, but would have sent one to: {0}".format(recips))


class Email(object):
    # Note: The docstring of an email is leveraged by code, hence this odd one.
    """
    (This email has no description.)
    """
    key = None
    fallback_key = None
    abstract = False
    default_prefix = "[rattail]"
    default_subject = "Automated message"

    def __init__(self, config, key=None, fallback_key=None):
        self.config = config

        if key:
            self.key = key
        elif not self.key:
            self.key = self.__class__.__name__
            if self.key == 'Email':
                raise exceptions.ConfigurationError("Email instance has no key: {0}".format(repr(self)))

        if fallback_key:
            self.fallback_key = fallback_key

        templates = config.getlist('rattail.mail', 'templates')
        templates = [resource_path(p) for p in templates]
        self.templates = TemplateLookup(directories=templates)

    def sample_data(self, request):
        return {}

    def get_enabled(self):
        """
        Get the enabled flag for the email's message type.
        """
        enabled = self.config.getbool('rattail.mail', '{0}.enabled'.format(self.key))
        if enabled is not None:
            return enabled
        enabled = self.config.getbool('rattail.mail', 'default.enabled')
        if enabled is not None:
            return enabled
        return self.config.getbool('rattail.mail', 'send_emails', default=True)

    def get_sender(self):
        """
        Returns the value for the message's ``From:`` header.

        :rtype: unicode
        """
        sender = self.config.get('rattail.mail', '{0}.from'.format(self.key))
        if not sender:
            sender = self.config.get('rattail.mail', 'default.from')
            if not sender:
                raise exceptions.SenderNotFound(self.key)
        return sender

    def get_replyto(self):
        """
        Get the Reply-To address for the message.
        """
        replyto = self.config.get('rattail.mail', '{0}.replyto'.format(self.key))
        if not replyto:
            replyto = self.config.get('rattail.mail', 'default.replyto')
        return replyto

    def get_recips(self, type_='to'):
        """
        Returns a list of recipients of the given type for the message.

        :param type_: Must be one of: ``('to', 'cc', 'bcc')``.

        :rtype: list
        """
        try:
            if type_.lower() not in ('to', 'cc', 'bcc'):
                raise Exception
        except:
            raise ValueError("Recipient type must be one of ('to', 'cc', 'bcc'); "
                             "not: {0}".format(repr(type_)))
        type_ = type_.lower()
        recips = self.config.getlist('rattail.mail', '{0}.{1}'.format(self.key, type_))
        if not recips:
            recips = self.config.getlist('rattail.mail', 'default.{0}'.format(type_))
        return recips

    def get_prefix(self, data={}):
        """
        Returns a string to be used as the subject prefix for the message.

        :rtype: unicode
        """
        prefix = self.config.get('rattail.mail', '{0}.prefix'.format(self.key))
        if not prefix:
            prefix = self.config.get('rattail.mail', 'default.prefix')
        return prefix or self.default_prefix

    def get_subject(self, data={}):
        """
        Returns the base value for the message's subject header, i.e. minus
        prefix.

        :rtype: unicode
        """
        subject = self.config.get('rattail.mail', '{0}.subject'.format(self.key),
                                  default=self.default_subject)
        if not subject:
            subject = self.config.get('rattail.mail', 'default.subject')
        if subject:
            subject = Template(subject).render(**data)
        return subject

    def get_complete_subject(self, data={}):
        """
        Returns the value for the message's ``Subject:`` header, i.e. the base
        subject with the prefix applied.  Note that config may provide the
        complete subject also, in which case the prefix and base subject are
        not considered.

        :rtype: unicode
        """
        return "{0} {1}".format(self.get_prefix(data), self.get_subject(data))

    def get_template(self, type_):
        """
        Locate and return the Mako email template of the given type
        (e.g. 'html'), or ``None`` if no such template can be found.
        """
        try:
            return self.templates.get_template('{0}.{1}.mako'.format(self.key, type_))
        except TopLevelLookupException:
            if self.fallback_key:
                try:
                    return self.templates.get_template('{0}.{1}.mako'.format(self.fallback_key, type_))
                except TopLevelLookupException:
                    pass

    def make_message(self, data={}, attachments=[]):
        """
        Returns a proper email ``Message`` instance which may be sent via SMTP.
        """
        txt_template = self.get_template('txt')
        html_template = self.get_template('html')

        if txt_template and html_template:
            msg = MIMEMultipart(_subtype='alternative', _subparts=[
                MIMEText(_text=txt_template.render(**data), _charset='utf_8'),
                MIMEText(_text=html_template.render(**data), _subtype='html', _charset='utf_8'),
            ])
            if attachments:
                msg = MIMEMultipart(_subtype='mixed', _subparts=[msg] + attachments)

        elif txt_template:
            if attachments:
                msg = MIMEMultipart(_subtype='mixed', _subparts=[
                    MIMEText(_text=txt_template.render(**data), _charset='utf_8'),
                ] + attachments)
            else:
                msg = MIMEText(_text=txt_template.render(**data), _charset='utf_8')

        elif html_template:
            if attachments:
                msg = MIMEMultipart(_subtype='mixed', _subparts=[
                    MIMEText(_text=html_template.render(**data), _subtype='html', _charset='utf_8'),
                ] + attachments)
            else:
                msg = MIMEText(_text=html_template.render(**data), _subtype='html', _charset='utf_8')

        else:
            raise exceptions.MailTemplateNotFound(self.key)

        # subject/from
        msg['Subject'] = self.get_complete_subject(data)
        msg['From'] = self.get_sender()

        # reply-to
        replyto = self.get_replyto()
        if replyto:
            msg.add_header('Reply-To', replyto)

        # recipients
        to = self.get_recips('to')
        cc = self.get_recips('cc')
        bcc = self.get_recips('bcc')
        if not (to or cc or bcc):
            raise exceptions.RecipientsNotFound(self.key)
        if to:
            for recip in to:
                msg['To'] = recip
        if cc:
            for recip in cc:
                msg['Cc'] = recip
        if bcc:
            for recip in bcc:
                msg['Bcc'] = recip

        return msg



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor invoice parser for Albert's Organics
"""

from __future__ import unicode_literals

import csv
import datetime

from rattail.vendors.invoices import InvoiceParser
from rattail.db.batch.vendorinvoice import VendorInvoiceRow
from rattail.gpc import GPC


class AlbertsInvoiceParser(InvoiceParser):
    """
    Invoice parser for Albert's Organics CSV files.
    """
    key = 'rattail.contrib.alberts'
    display = "Albert's Organics"
    vendor_key = 'alberts'

    def parse_invoice_date(self, path):
        with open(path, 'rb') as f:
            reader = csv.DictReader(f)
            data = reader.next()
            return datetime.datetime.strptime(data['Invoice Date'], '%m/%d/%Y').date()

    def parse_rows(self, path):
        with open(path, 'rb') as f:
            reader = csv.DictReader(f)
            for data in reader:

                row = VendorInvoiceRow()
                row.upc = GPC(data['UPCPLU'].replace('-', '').replace(' ', ''))
                row.vendor_code = data['ItemNumber']
                row.brand_name = data['BrandName']
                row.description = data['MV']
                row.size = "{0} {1}".format(data['PkgSize'], data['UOMAbbr'])
                row.case_quantity = self.int_(data['PkgCount']) or int(self.decimal(data['PkgSize']))
                row.shipped_cases = self.decimal(data['ShipQty'])
                row.case_cost = self.decimal(data['CasePrice'])
                row.unit_cost = self.decimal(data['EachPrice'])
                row.total_cost = row.case_cost * row.shipped_cases

                yield row



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Time Utilities
"""

from __future__ import unicode_literals

import datetime
import warnings

import pytz


def localtime(config, time=None, key='default', tzinfo=True):
    """
    Return a datetime which has been localized to a particular timezone.  The
    :func:`timezone()` function will be used to obtain the timezone to which
    the time value will be localized.

    :param config: Reference to a config object.

    :param time: Optional :class:`python:datetime.datetime` instance to be
       localized.  If not provided, the current time ("now") is assumed.

    :param tzinfo: Boolean indicating whether the result should contain
       ``tzinfo`` or not, i.e. whether it should be time zone "aware"
       (``True``) or "naive" (``False``).

    :param key: Config key to be used in determining to which timezone the time
       should be localized.
    """
    zone = timezone(config, key)
    if time is None:
        time = datetime.datetime.utcnow()
        time = pytz.utc.localize(time)
        time = zone.normalize(time.astimezone(zone))
    elif time.tzinfo:
        time = zone.normalize(time.astimezone(zone))
    else:
        time = zone.localize(time)
    if not tzinfo:
        time = time.replace(tzinfo=None)
    return time


def timezone(config, key='default'):
    """
    Return a timezone object based on the definition found in config.

    :param config: Reference to a config object.

    :param key: Config key used to determine which timezone should be returned.

    :returns: A :class:`pytz:pytz.tzinfo` instance, created using the Olson
       time zone name found in the config file.

    An example of the configuration syntax which is assumed by this function:

    .. code-block:: ini

       [rattail]

       # retrieve with: timezone(config)
       timezone.default = America/Los_Angeles

       # retrieve with: timezone(config, 'headoffice')
       timezone.headoffice = America/Chicago

       # retrieve with: timezone(config, 'satellite')
       timezone.satellite = America/New_York

    See `Wikipedia`_ for the full list of Olson time zone names.

    .. _`Wikipedia`: http://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """
    # Don't *require* the correct config option just yet, so we can fall back
    # to 'local' for default if necessary.
    zone = config.get('rattail', 'timezone.{}'.format(key), usedb=False)
    if zone is None and key == 'default':
        zone = config.get('rattail', 'timezone.local', usedb=False)
    if zone is None:
        # Okay, now let's require the correct one.
        zone = config.require('rattail', 'timezone.{0}'.format(key))
    return pytz.timezone(zone)


def make_utc(time):
    """
    Convert a timezone-aware time back to a naive UTC equivalent.
    """
    utctime = pytz.utc.normalize(time.astimezone(pytz.utc))
    return utctime.replace(tzinfo=None)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Customers
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.orderinglist import ordering_list

from .core import Base, uuid_column, getset_factory
from .contact import PhoneNumber, EmailAddress
from .people import Person


class Customer(Base):
    """
    Represents a customer account.

    Customer accounts may consist of more than one person, in some cases.
    """
    __tablename__ = 'customer'
    __versioned__ = {}

    uuid = uuid_column()
    id = sa.Column(sa.String(length=20))
    name = sa.Column(sa.String(length=255))
    email_preference = sa.Column(sa.Integer())

    def __unicode__(self):
        return unicode(self.name or self.person)

    def add_email_address(self, address, type='Home'):
        email = CustomerEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Home'):
        phone = CustomerPhoneNumber(number=number, type=type)
        self.phones.append(phone)


class CustomerPhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Customer`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Customer'}


Customer.phones = orm.relationship(
    CustomerPhoneNumber,
    backref='customer',
    primaryjoin=CustomerPhoneNumber.parent_uuid == Customer.uuid,
    foreign_keys=[CustomerPhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=CustomerPhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Customer.phone = orm.relationship(
    CustomerPhoneNumber,
    primaryjoin=sa.and_(
        CustomerPhoneNumber.parent_uuid == Customer.uuid,
        CustomerPhoneNumber.preference == 1),
    foreign_keys=[CustomerPhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class CustomerEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Customer`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Customer'}


Customer.emails = orm.relationship(
    CustomerEmailAddress,
    backref='customer',
    primaryjoin=CustomerEmailAddress.parent_uuid == Customer.uuid,
    foreign_keys=[CustomerEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=CustomerEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Customer.email = orm.relationship(
    CustomerEmailAddress,
    primaryjoin=sa.and_(
        CustomerEmailAddress.parent_uuid == Customer.uuid,
        CustomerEmailAddress.preference == 1),
    foreign_keys=[CustomerEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)


class CustomerGroup(Base):
    """
    Represents an arbitrary group to which customers may belong.
    """
    __tablename__ = 'customer_group'
    __versioned__ = {}

    uuid = uuid_column()
    id = sa.Column(sa.String(length=20))
    name = sa.Column(sa.String(length=255))

    def __unicode__(self):
        return unicode(self.name or '')


class CustomerGroupAssignment(Base):
    """
    Represents the assignment of a customer to a group.
    """
    __tablename__ = 'customer_x_group'
    __table_args__ = (
        sa.ForeignKeyConstraint(['group_uuid'], ['customer_group.uuid'], name='customer_x_group_fk_group'),
        sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'], name='customer_x_group_fk_customer'),
        )

    uuid = uuid_column()
    customer_uuid = sa.Column(sa.String(length=32), nullable=False)
    group_uuid = sa.Column(sa.String(length=32), nullable=False)
    ordinal = sa.Column(sa.Integer(), nullable=False)

    group = orm.relationship(CustomerGroup)


Customer._groups = orm.relationship(
    CustomerGroupAssignment, backref='customer',
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=CustomerGroupAssignment.ordinal,
    cascade='save-update, merge, delete, delete-orphan')

Customer.groups = association_proxy(
    '_groups', 'group',
    getset_factory=getset_factory,
    creator=lambda g: CustomerGroupAssignment(group=g))


class CustomerPerson(Base):
    """
    Represents the association between a person and a customer account.
    """
    __tablename__ = 'customer_x_person'
    __table_args__ = (
        sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'], name='customer_x_person_fk_customer'),
        sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='customer_x_person_fk_person'),
        )

    uuid = uuid_column()
    customer_uuid = sa.Column(sa.String(length=32), nullable=False)
    person_uuid = sa.Column(sa.String(length=32), nullable=False)
    ordinal = sa.Column(sa.Integer(), nullable=False)

    customer = orm.relationship(Customer, back_populates='_people')

    person = orm.relationship(Person, back_populates='_customers')


Customer._people = orm.relationship(
    CustomerPerson, back_populates='customer',
    primaryjoin=CustomerPerson.customer_uuid == Customer.uuid,
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=CustomerPerson.ordinal,
    cascade='save-update, merge, delete, delete-orphan')

Customer.people = association_proxy(
    '_people', 'person',
    getset_factory=getset_factory,
    creator=lambda p: CustomerPerson(person=p))

Customer._person = orm.relationship(
    CustomerPerson,
    primaryjoin=sa.and_(
        CustomerPerson.customer_uuid == Customer.uuid,
        CustomerPerson.ordinal == 1),
    uselist=False,
    viewonly=True)

Customer.person = association_proxy(
    '_person', 'person',
    getset_factory=getset_factory)

Person._customers = orm.relationship(
    CustomerPerson, back_populates='person',
    primaryjoin=CustomerPerson.person_uuid == Person.uuid,
    cascade='all, delete-orphan',
    viewonly=True)

Person.customers = association_proxy('_customers', 'customer',
                                     getset_factory=getset_factory,
                                     creator=lambda c: CustomerPerson(customer=c))



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2014 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
SIL Columns
"""

from __future__ import unicode_literals

from ..core import Object
from .exceptions import SILColumnNotFound
from ..util import load_entry_points


__all__ = ['get_column']


supported_columns = None


class SILColumn(Object):
    """
    Represents a column for use with SIL.
    """

    def __init__(self, name, data_type, description, display_name=None, **kwargs):
        Object.__init__(self, **kwargs)
        self.name = name
        self.data_type = data_type
        self.description = description
        self.display_name = display_name or description

    def __repr__(self):
        return "<SILColumn: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name)


def provide_columns():
    """
    Provides all SIL columns natively supported by Rattail.
    """

    SC = SILColumn

    standard = [ # These columns are part of the SIL standard.

        # ITEM_DCT
        SC('F01',       'GPC(14)',      "Primary Item U.P.C. Number (Key)",     "UPC"),
        SC('F02',       'CHAR(20)',     "Descriptor",                           "Description"),
        SC('F04',       'NUMBER(4,0)',  "Sub-Department Number"),
        SC('F22',       'CHAR(30)',     "Size Description"),
        SC('F26',       'CHAR(15)',     "Primary Item Code (User)"),
        SC('F90',       'FLAG(1)',      "Authorized DSD Item"),
        SC('F94',       'NUMBER(2,0)',  "Shelf Tag Quantity"),
        SC('F95',       'CHAR(3)',      "Shelf Tag Type"),
        SC('F155',      'CHAR(30)',     "Brand"),
        SC('F188',      'FLAG(1)',      "POS Valid Item"),

        # PRICE_DCT
        SC('F30',       'NUMBER(8,3)',  "Retail Sell Price"),
        SC('F31',       'NUMBER(3,0)',  "Price Multiple (Quantity/For)"),
        SC('F35',       'DATE(7)',      "Price Start Date"),
        SC('F36',       'TIME(4)',      "Price Start Time"),
        SC('F126',      'NUMBER(2,0)',  "Pricing Level"),
        SC('F129',      'DATE(7)',      "Price End Date"),
        SC('F130',      'TIME(4)',      "Price End Time"),
        SC('F135',      'NUMBER(3,0)',  "Sale Price Multiple (Quantity/For)"),
        SC('F136',      'NUMBER(8,3)',  "Sale Price"),
        SC('F137',      'DATE(7)',      "Sale Price Start Date"),
        SC('F138',      'TIME(4)',      "Sale Price End Date"),
        SC('F139',      'NUMBER(8,3)',  "Sale Package Price"),
        SC('F140',      'NUMBER(8,3)',  "Package Price"),
        SC('F142',      'NUMBER(3,0)',  "Package Price Multiple"),
        SC('F143',      'NUMBER(3,0)',  "Sale Package Price Multiple"),
        SC('F144',      'TIME(4)',      "Sale Price Start Time"),
        SC('F145',      'TIME(4)',      "Sale Price End Time"),
        SC('F181',      'NUMBER(8,3)',  "TPR (Temporary Price Reduction)"),
        SC('F182',      'NUMBER(3,0)',  "TPR Multiple (Quantity/For)"),
        SC('F183',      'DATE(7)',      "TPR Start Date"),
        SC('F184',      'DATE(7)',      "TPR End Date"),
        SC('F387',      'NUMBER(3)',    "Price Type Code"),

        # FCOST_DCT
        SC('F19',       'NUMBER(4,0)',  "Case Pack Size"),
        SC('F20',       'NUMBER(4,0)',  "Receiving Pack Size"),
        SC('F38',       'NUMBER(9,5)',  "Case Receiving Base Cost"),
        SC('F212',      'TIME(4)',      "Cost Change Time"),
        SC('F227',      'DATE(7)',      "Cost Change Date"),

        # DEPT_DCT
        SC('F03',       'NUMBER(4,0)',  "Department Number"),
        SC('F238',      'CHAR(30)',     "Department Description"),

        # VENDOR_DCT
        SC('F27',       'CHAR(9)',      "Vendor Number"),
        SC('F334',      'CHAR(20)',     "Vendor Name"),
        SC('F335',      'CHAR(20)',     "Vendor Contact Name"),
        SC('F341',      'NUMBER(10,0)', "Vendor Phone Number - Voice"),
        SC('F342',      'NUMBER(10,0)', "Vendor Phone Number - Fax"),
        ]

    custom = [ # These columns are Rattail-specific.

        SC('R38',       'NUMBER(9,5)',  "Unit Receiving Base Cost"),
        SC('R49',       'NUMBER(5,3)',  "Gross Margin Percent"),
        SC('R71',       'NUMBER(5,0)',  "Inventory - Units Counted",            "Units"),
        SC('R72',       'NUMBER(5,0)',  "Inventory - Cases Counted",            "Cases"),
        SC('R101',      'NUMBER(9,5)',  "Difference Amount",                    "Difference"),
        SC('R102',      'NUMBER(3,0)',  "Status Code"),
        SC('R103',      'CHAR(30)',     "Status Text"),
        ]

    columns = {}
    for column in standard + custom:
        columns[column.name] = column
    return columns


def get_column(name):
    """
    Returns the :class:`SILColumn` instance named ``name``.
    """

    global supported_columns

    if supported_columns is None:
        supported_columns = {}
        providers = load_entry_points('rattail.sil.column_providers')
        for provider in providers.itervalues():
            supported_columns.update(provider())

    column = supported_columns.get(name)
    if not column:
        raise SILColumnNotFound(name)
    return column



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for Vendor Catalog batches
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy.orm import joinedload

from rattail.db import model
from rattail.db.batch.handler import FileBatchHandler
from rattail.vendors.catalogs import require_catalog_parser


class VendorCatalogHandler(FileBatchHandler):
    """
    Handler for vendor catalog batches.
    """
    batch_model_class = model.VendorCatalog
    show_progress = True

    def refresh_data(self, session, batch, progress=None):
        """
        Refresh all data for the batch.
        """
        del batch.data_rows[:]
        data_path = batch.filepath(self.config)
        parser = require_catalog_parser(batch.parser_key)
        parser.session = session
        parser.vendor = batch.vendor
        batch.effective = parser.parse_effective_date(data_path)

        # Pre-cache products by UPC and vendor code.
        self.vendor = parser.vendor
        self.products = {'upc': {}, 'vendor_code': {}}
        products = session.query(model.Product)\
            .options(joinedload(model.Product.brand))\
            .options(joinedload(model.Product.costs))
        prog = None
        if progress:
            prog = progress("Caching products by UPC and vendor item code", products.count())
        for i, product in enumerate(products, 1):
            if product.upc:
                self.products['upc'][product.upc] = product
            cost = product.cost_for_vendor(self.vendor)
            product.vendor_cost = cost
            if cost and cost.code:
                self.products['vendor_code'][cost.code] = product
            if prog:
                prog.update(i)
        if prog:
            prog.destroy()

        # Get data from parser, and convert to rows.
        data = list(parser.parse_rows(data_path))
        self.make_rows(session, batch, data, progress=progress)

    def cognize_row(self, session, row):
        """
        Inspect a single row from a catalog, and set its attributes based on
        whether or not the product exists, if we already have a cost record for
        the vendor, if the catalog contains a change etc.  Note that the
        product lookup is done first by UPC and then by vendor item code.
        """
        if row.upc:
            row.product = self.products['upc'].get(row.upc)
        if not row.product and row.vendor_code:
            row.product = self.products['vendor_code'].get(row.vendor_code)
        if not row.product:
            row.status_code = row.STATUS_PRODUCT_NOT_FOUND
            return

        row.upc = row.product.upc
        row.brand_name = row.product.brand.name if row.product.brand else None
        row.description = row.product.description
        row.size = row.product.size

        old_cost = row.product.vendor_cost
        if not old_cost:
            row.status_code = row.STATUS_NEW_COST
            return

        row.old_vendor_code = old_cost.code
        row.old_case_size = old_cost.case_size

        row.old_case_cost = old_cost.case_cost
        if row.case_cost is not None and row.old_case_cost is not None:
            row.case_cost_diff = row.case_cost - row.old_case_cost

        row.old_unit_cost = old_cost.unit_cost
        if row.unit_cost is not None and row.old_unit_cost is not None:
            row.unit_cost_diff = row.unit_cost - row.old_unit_cost

        diff = self.cost_differs(row, old_cost)
        if diff:
            row.status_code = row.STATUS_UPDATE_COST
            row.status_text = diff
        else:
            row.status_code = row.STATUS_NO_CHANGE

    def cost_differs(self, row, cost):
        """
        Compare a batch row with a cost record to determine whether they match
        or differ.
        """
        if row.vendor_code is not None and row.vendor_code != cost.code:
            return "new vendor code {} differs from old code {}".format(
                repr(row.vendor_code), repr(cost.code))
        if row.case_cost is not None and row.case_cost != cost.case_cost:
            return "new case cost {} differs from old cost {}".format(
                row.case_cost, cost.case_cost)
        if row.unit_cost is not None and row.unit_cost != cost.unit_cost:
            return "new unit cost {} differs from old cost {}".format(
                row.unit_cost, cost.unit_cost)

    def execute(self, batch, progress=None):
        return True



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2014 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
**Package Root**

The only things present in the ``rattail`` root namespace are:

.. attribute:: __version__

   A string indicating the version of the Rattail core package.

.. attribute:: enum
   :noindex:

   The :mod:`rattail.enum` module, imported at load time for convenience.
"""

from ._version import __version__
from . import enum

# TODO: Remove this at some point, when callers are ready.  Maybe would be nice
# to deprecate somehow in the meantime?
from rattail.enum import *



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Importers
"""

from __future__ import unicode_literals, absolute_import

import datetime
import logging

from sqlalchemy import orm
from sqlalchemy.orm.exc import NoResultFound

from rattail.db import cache


log = logging.getLogger(__name__)


class Importer(object):
    """
    Base class for all data importers.
    """
    key = 'uuid'
    normalize_progress_message = "Normalizing source data"
    cached_instances = None

    def __init__(self, config=None, session=None, fields=None, key=None, **kwargs):
        self.config = config
        self.session = session
        self.fields = fields or self.supported_fields
        if key:
            self.key = key
        if isinstance(self.key, basestring):
            self.key = (self.key,)
        for key, value in kwargs.iteritems():
            setattr(self, key, value)

    @property
    def model_class(self):
        """
        This should return a reference to the model class which the importer
        "targets" so to speak.
        """
        raise NotImplementedError

    @property
    def model_name(self):
        """
        Returns the string 'name' of the model class which the importer targets.
        """
        return self.model_class.__name__

    @property
    def simple_fields(self):
        """
        The list of field names which may be considered "simple" and therefore
        treated as such, i.e. with basic getattr/setattr calls.  Note that this
        only applies to the local / target side, it has no effect on the
        upstream / foreign side.
        """
        mapper = orm.class_mapper(self.model_class)
        return list(mapper.columns.keys())

    @property
    def supported_fields(self):
        """
        The list of field names which are supported in general by the importer.
        Note that this only applies to the local / target side, it has no
        effect on the upstream / foreign side.
        """
        return self.simple_fields

    def setup(self):
        """
        Perform any setup necessary, e.g. cache lookups for existing data.
        """

    def teardown(self):
        """
        Perform any cleanup after import, if necessary.
        """

    def import_data(self, args, progress=None):
        """
        Import some data!  This is the core body of logic for that, regardless
        of where data is coming from or where it's headed.  Note that this
        method handles deletions as well as adds/updates.
        """
        self.now = datetime.datetime.utcnow()
        self.progress = progress
        self.setup()
        created = updated = deleted = []

        data = self.normalize_source_data()

        self.cached_instances = self.cache_instance_data()

        # Normalize source data set in order to prune duplicate keys.  This is
        # for the sake of sanity since duplicates typically lead to a ping-pong
        # effect, where a "clean" (change-less) import is impossible.
        unique = {}
        for record in data:
            key = self.get_key(record)
            if key in unique:
                log.warning("duplicate records detected from {}:{} for key: {}".format(
                    self.__class__.__module__, self.__class__.__name__, key))
            unique[key] = record
        data = []
        for key in sorted(unique):
            data.append(unique[key])

        created = updated = deleted = []

        if args.create or args.update:
            created, updated = self._import_create_update(data, args)

        if args.delete:
            changes = len(created) + len(updated)
            if args.max_total and changes >= args.max_total:
                log.warning("max of {} total changes already reached; skipping deletions".format(args.max_total))
            else:
                deleted = self._import_delete(data, args, host_keys=set(unique), changes=changes)

        self.teardown()
        return created, updated, deleted

    def _import_create_update(self, data, args):
        """
        Import the given data; create and/or update records as needed and
        according to the args provided.
        """
        created = []
        updated = []
        count = len(data)
        if not count:
            return created, updated

        prog = None
        if self.progress:
            prog = self.progress("Importing {} data".format(self.model_name), count)

        keys_seen = set()
        for i, source_data in enumerate(data, 1):

            # Get what should be the unique key for the current 'host' data
            # record, but warn if we find it to be not unique.  Note that we
            # will still wind up importing both records however.
            key = self.get_key(source_data)
            if key in keys_seen:
                log.warning("duplicate records from {}:{} for key: {}".format(
                        self.__class__.__module__, self.__class__.__name__, key))
            else:
                keys_seen.add(key)

            # Fetch local instance, using key from host record.
            instance = self.get_instance(key)

            # If we have a local instance, but its data differs from host, update it.
            if instance and args.update:
                instance_data = self.normalize_instance(instance)
                diffs = self.data_diffs(instance_data, source_data)
                if diffs:
                    log.debug("fields '{}' differed for local data: {}, host data: {}".format(
                            ','.join(diffs), instance_data, source_data))
                    self.update_instance(instance, source_data, instance_data)
                    self.session.flush()
                    updated.append(instance)
                    if args.max_update and len(updated) >= args.max_update:
                        log.warning("max of {} *updated* records has been reached; stopping now".format(args.max_update))
                        break
                    if args.max_total and (len(created) + len(updated)) >= args.max_total:
                        log.warning("max of {} *total changes* has been reached; stopping now".format(args.max_total))
                        break

            # If we did not yet have a local instance, create it using host data.
            elif not instance and args.create:
                instance = self.new_instance(key)
                self.update_instance(instance, source_data)
                self.session.add(instance)
                self.session.flush()
                log.debug("created new {} {}: {}".format(self.model_name, key, instance))
                created.append(instance)
                if self.cached_instances is not None:
                    self.cached_instances[key] = {'instance': instance, 'data': self.normalize_instance(instance)}
                if args.max_create and len(created) >= args.max_create:
                    log.warning("max of {} *created* records has been reached; stopping now".format(args.max_create))
                    break
                if args.max_total and (len(created) + len(updated)) >= args.max_total:
                    log.warning("max of {} *total changes* has been reached; stopping now".format(args.max_total))
                    break

            if prog:
                prog.update(i)
        if prog:
            prog.destroy()

        return created, updated

    def get_deletion_keys(self):
        """
        Return a set of keys from the *local* data set, which are eligible for
        deletion.  By default this will be all keys from the local (cached)
        data set.
        """
        return set(self.cached_instances)

    def _import_delete(self, data, args, host_keys=None, changes=0):
        """
        Import deletions for the given data set.
        """
        if host_keys is None:
            host_keys = set([self.get_key(rec) for rec in data])

        deleted = []
        deleting = self.get_deletion_keys() - host_keys
        count = len(deleting)
        log.debug("found {} instances to delete".format(count))
        if count:

            prog = None
            if self.progress:
                prog = self.progress("Deleting {} data".format(self.model_name), count)

            for i, key in enumerate(sorted(deleting), 1):

                instance = self.cached_instances.pop(key)['instance']
                if self.delete_instance(instance):
                    deleted.append(instance)

                    if args.max_delete and len(deleted) >= args.max_delete:
                        log.warning("max of {} *deleted* records has been reached; stopping now".format(args.max_delete))
                        break
                    if args.max_total and (changes + len(deleted)) >= args.max_total:
                        log.warning("max of {} *total changes* has been reached; stopping now".format(args.max_total))
                        break

                if prog:
                    prog.update(i)
            if prog:
                prog.destroy()

        return deleted

    def delete_instance(self, instance):
        """
        Process a deletion for the given instance.  The default implementation
        really does delete the instance from the local session, so you must
        override this if you need something else to happen.

        This method must return a boolean indicating whether or not the
        deletion was performed.  This implies for example that you may simply
        do nothing, and return ``False``, to effectively disable deletion
        altogether for an importer.
        """
        self.session.delete(instance)
        self.session.flush()
        self.session.expunge(instance)
        return True

    def get_source_data(self):
        """
        Return the "raw" (as-is, not normalized) data which is to be imported.
        This may be any sequence-like object, which has a ``len()`` value and
        responds to iteration etc.  The objects contained within it may be of
        any type, no assumptions are made there.  (That is the job of the
        :meth:`normalize_source_data()` method.)
        """
        return []

    def normalize_source_data(self):
        """
        Return a normalized version of the full set of source data.  Note that
        this calls :meth:`get_source_data()` to obtain the initial data set,
        and then normalizes each record.  the normalization process may filter
        out some records from the set, in which case the return value will be
        smaller than the original data set.
        """
        source_data = self.get_source_data()
        normalized = []
        count = len(source_data)
        if count == 0:
            return normalized
        prog = None
        if self.progress:
            prog = self.progress(self.normalize_progress_message, count)
        for i, data in enumerate(source_data, 1):
            data = self.normalize_source_record(data)
            if data:
                normalized.append(data)
            if prog:
                prog.update(i)
        if prog:
            prog.destroy()
        return normalized

    def get_key(self, data):
        """
        Return the key value for the given data record.
        """
        return tuple(data[k] for k in self.key)

    def int_(self, value):
        """
        Coerce ``value`` to an integer, or return ``None`` if that can't be
        done cleanly.
        """
        try:
            return int(value)
        except (TypeError, ValueError):
            return None

    def prioritize_2(self, data, field):
        """
        Prioritize the data values for the pair of fields implied by the given
        fieldname.  I.e., if only one non-empty value is present, make sure
        it's in the first slot.
        """
        field2 = '{}_2'.format(field)
        if field in data and field2 in data:
            if data[field2] and not data[field]:
                data[field], data[field2] = data[field2], None

    def normalize_source_record(self, record):
        """
        Normalize a source data record.  Generally this is where the importer
        may massage the record in any way necessary, so that its values are
        more "native" and can be used for direct comparison with, and
        assignment to, the target model instance.

        Note that if you override this, your method must return the data to be
        imported.  If your method returns ``None`` then that particular record
        would be skipped and not imported.
        """
        return record

    def cache_model(self, model, **kwargs):
        """
        Convenience method which invokes :func:`rattail.db.cache.cache_model()`
        with the given model and keyword arguments.  It will provide the
        ``session`` and ``progress`` parameters by default, setting them to the
        importer's attributes of the same names.
        """
        session = kwargs.pop('session', self.session)
        kwargs.setdefault('progress', self.progress)
        return cache.cache_model(session, model, **kwargs)

    def cache_instance_data(self):
        """
        Cache all existing model instances as normalized data.
        """
        return cache.cache_model(self.session, self.model_class,
                                 key=self.get_cache_key,
                                 omit_duplicates=True,
                                 query_options=self.cache_query_options(),
                                 normalizer=self.normalize_cache_instance,
                                 progress=self.progress)

    def cache_query_options(self):
        """
        Return a list of options to apply to the cache query, if needed.
        """

    def get_cache_key(self, instance, normalized):
        """
        Get the primary model cache key for a given instance/data object.
        """
        return tuple(normalized['data'].get(k) for k in self.key)

    def normalize_cache_instance(self, instance):
        """
        Normalizer for cache data.  This adds the instance to the cache in
        addition to its normalized data.  This is so that if lots of updates
        are required, we don't we have to constantly fetch them.
        """
        return {'instance': instance, 'data': self.normalize_instance(instance)}

    def get_instance(self, key):
        """
        Fetch an instance from our database which corresponds to the source
        data, if possible; otherwise return ``None``.
        """
        if self.cached_instances:
            data = self.cached_instances.get(key)
            return data['instance'] if data else None

        query = self.session.query(self.model_class)
        for i, k in enumerate(self.key):
            query = query.filter(getattr(self.model_class, k) == key[i])

        try:
            return query.one()
        except NoResultFound:
            pass

    def normalize_instance(self, instance):
        """
        Normalize a model instance.
        """
        data = {}
        for field in self.simple_fields:
            if field in self.fields:
                data[field] = getattr(instance, field)
        return data

    def newval(self, data, field, value):
        """
        Assign a "new" field value to the given data record.  In other words
        don't try to be smart about not overwriting it if the existing data
        already matches etc.  However the main point of this is to skip fields
        which are not included in the current task.
        """
        if field in self.fields:
            data[field] = value

    def data_diffs(self, model_data, host_data):
        """
        Find all (relevant) fields which differ between the model and host data
        for a given record.
        """
        diffs = []
        for field in self.fields:
            if model_data[field] != host_data[field]:
                diffs.append(field)
        return diffs

    def new_instance(self, key):
        """
        Return a new model instance to correspond to the given key.
        """
        instance = self.model_class()
        for i, k in enumerate(self.key):
            setattr(instance, k, key[i])
        return instance

    def update_instance(self, instance, data, instance_data=None):
        """
        Update the given model instance with the given data.
        """
        for field in self.simple_fields:
            if field in self.fields:
                if not instance_data or instance_data[field] != data[field]:
                    setattr(instance, field, data[field])


class QueryDataProxy(object):
    """
    Simple proxy to wrap a SQLAlchemy (or Django) query and make it sort of
    behave like a normal sequence, as much as needed to make an importer happy.
    """

    def __init__(self, query):
        self.query = query

    def __len__(self):
        return self.query.count()

    def __iter__(self):
        return iter(self.query)


class QueryImporter(Importer):
    """
    Base class for importers whose raw external data source is a SQLAlchemy (or
    Django) query.
    """

    def query(self):
        """
        Must return the primary query which will define the data set.
        """
        raise NotImplementedError

    def get_source_data(self, progress=None):
        return QueryDataProxy(self.query())


class SQLAlchemyImporter(QueryImporter):
    """
    Base class for importers whose external data source is a SQLAlchemy query.
    """
    host_session = None

    @property
    def host_model_class(self):
        """
        For default behavior, set this to a model class to be used in
        generating the host (source) data query.
        """
        raise NotImplementedError

    def query(self):
        """
        Must return the primary query which will define the data set.  Default
        behavior is to leverage :attr:`host_session` and generate a query for
        the class defined by :attr:`host_model_class`.
        """
        return self.host_session.query(self.host_model_class)



# -*- coding: utf-8 -*-
"""add vendor_invoice

Revision ID: 218562884128
Revises: 268ff6454410
Create Date: 2015-02-13 21:51:41.779081

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '218562884128'
down_revision = u'268ff6454410'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():
    op.create_table('vendor_invoice',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.Column('filename', sa.String(length=255), nullable=False),
                    sa.Column('parser_key', sa.String(length=100), nullable=False),
                    sa.Column('vendor_uuid', sa.String(length=32), nullable=False),
                    sa.Column('invoice_date', sa.Date(), nullable=True),
                    sa.Column('purchase_order_number', sa.String(length=20), nullable=True),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], [u'user.uuid'], name=u'vendor_invoice_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'vendor_invoice_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], [u'user.uuid'], name=u'vendor_invoice_fk_executed_by'),
                    sa.ForeignKeyConstraint(['vendor_uuid'], [u'vendor.uuid'], name=u'vendor_invoice_fk_vendor'),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_table('vendor_invoice_row',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('vendor_code', sa.String(length=30), nullable=True),
                    sa.Column('case_quantity', sa.Integer(), nullable=False),
                    sa.Column('ordered_cases', sa.Integer(), nullable=True),
                    sa.Column('ordered_units', sa.Integer(), nullable=True),
                    sa.Column('shipped_cases', sa.Integer(), nullable=True),
                    sa.Column('shipped_units', sa.Integer(), nullable=True),
                    sa.Column('case_cost', sa.Numeric(precision=9, scale=5), nullable=True),
                    sa.Column('unit_cost', sa.Numeric(precision=9, scale=5), nullable=True),
                    sa.Column('total_cost', sa.Numeric(precision=9, scale=5), nullable=True),
                    sa.Column('line_number', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], [u'vendor_invoice.uuid'], name=u'vendor_invoice_row_fk_batch'),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'vendor_invoice_row_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
                    )


def downgrade():
    op.drop_table('vendor_invoice_row')
    op.drop_table('vendor_invoice')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2014 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
People API
"""

from __future__ import unicode_literals

from sqlalchemy.orm.exc import NoResultFound

from rattail.db import model


__all__ = ['get_employee_by_id']


def get_employee_by_id(session, employee_id):
    """
    Fetch an employee record by ID.
    """

    try:
        return session.query(model.Employee)\
            .filter(model.Employee.id == employee_id)\
            .one()
    except NoResultFound:
        return None



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Import Handlers
"""

from __future__ import unicode_literals, absolute_import

import logging

from rattail.util import OrderedDict
from rattail.mail import send_email


log = logging.getLogger(__name__)


class ImportHandler(object):
    """
    Base class for all import handlers.
    """

    def __init__(self, config=None, session=None):
        self.config = config
        self.session = session
        self.importers = self.get_importers()

    def get_importers(self):
        """
        Returns a dict of all available importers, where the values are
        importer factories.  All subclasses will want to override this.  Note
        that if you return an ``OrderedDict`` instance, you can affect the
        ordering of keys in the command line help system, etc.
        """
        return {}

    def get_importer_keys(self):
        """
        Returns a list of keys corresponding to the available importers.
        """
        return list(self.importers.iterkeys())

    def get_importer(self, key):
        """
        Returns an importer instance corresponding to the given key.
        """
        return self.importers[key](self.config, self.session,
                                   **self.get_importer_kwargs(key))

    def get_importer_kwargs(self, key):
        """
        Return a dict of kwargs to be used when construcing an importer with
        the given key.
        """
        return {}

    def import_data(self, keys, max_updates=None, progress=None):
        """
        Import all data for the given importer keys.
        """
        self.before_import()
        updates = OrderedDict()

        for key in keys:
            provider = self.get_importer(key)
            if not provider:
                log.warning("unknown importer; skipping: {0}".format(repr(key)))
                continue

            data = provider.get_data(progress=progress)
            created, updated = provider.importer.import_data(
                data, provider.supported_fields, provider.key,
                max_updates=max_updates, progress=progress)

            if hasattr(provider, 'process_deletions'):
                deleted = provider.process_deletions(data, progress=progress)
            else:
                deleted = 0

            log.info("added {0}, updated {1}, deleted {2} {3} records".format(
                len(created), len(updated), deleted, key))
            if created or updated or deleted:
                updates[key] = created, updated, deleted

        self.after_import()
        return updates

    def before_import(self):
        return 

    def after_import(self):
        return 

    def process_warnings(self, updates, command=None, **kwargs):
        """
        If an import was run with "warnings" enabled, and work was effectively
        done then this method is called to process the updates.  The assumption
        is that a warning email will be sent with the details, but you can do
        anything you like if you override this.
        """
        data = kwargs
        data['updates'] = updates

        if command:
            data['command'] = '{} {}'.format(command.parent.name, command.name)
        else:
            data['command'] = None

        if command:
            key = '{}_{}_updates'.format(command.parent.name, command.name)
            key = key.replace('-', '_')
        else:
            key = 'rattail_import_updates'

        send_email(self.config, key, fallback_key='rattail_import_updates', data=data)



# -*- coding: utf-8 -*-
"""make invoice quantities decimal

Revision ID: 3ef0fbcfb42e
Revises: 2dd756a44a16
Create Date: 2015-02-23 20:02:57.049418

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3ef0fbcfb42e'
down_revision = u'2dd756a44a16'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():
    op.alter_column('vendor_invoice_row', 'ordered_cases', type_=sa.Numeric(precision=9, scale=4))
    op.alter_column('vendor_invoice_row', 'ordered_units', type_=sa.Numeric(precision=9, scale=4))
    op.alter_column('vendor_invoice_row', 'shipped_cases', type_=sa.Numeric(precision=9, scale=4))
    op.alter_column('vendor_invoice_row', 'shipped_units', type_=sa.Numeric(precision=9, scale=4))


def downgrade():
    op.alter_column('vendor_invoice_row', 'ordered_cases', type_=sa.Integer())
    op.alter_column('vendor_invoice_row', 'ordered_units', type_=sa.Integer())
    op.alter_column('vendor_invoice_row', 'shipped_cases', type_=sa.Integer())
    op.alter_column('vendor_invoice_row', 'shipped_units', type_=sa.Integer())
